/* sim.c   run robot simulation 
 * 29 Jan 96
 * 9 Feb 96   change : precompute hitq[]
 * 7 Mar 96
 * simplified 5 Mar 97
 * P. Chongstitvatana
 */

#include "gp.h"

char  im[IMW][IMH];				/* grid of working area  */

char  finishF,dieF,hitF,outF,incF,decF,outF,seeF;  /* simulator flags */
int   pain;
int   ds,de,dw;					/* accumulated motions from each evalprog() */

point shoulder, elbow, wrist, tip, target;     /* absolute position */
int   zshoulder,zelbow,zwrist;	/* absolute angles, 0..TS-1, TE-1, TW-1 */
int   sumdist;
int	  NearGoal;					/* radius around target ^ 2 */	

sethome() /* initialise robot parameters */
{
  zshoulder = 20;
  zelbow = 26;
  zwrist = 17;
  shoulder.x = SX;
  shoulder.y = SY;
  armpos();
  hitF = seeF = incF = decF = outF = 0;
}

/* return 0..5 indicates the part of the arm that hit, 6 not hit
	    0      2        4  
	s ----- e ----- w ----- t
        1      3        5
 */
PRIVATE int ishit() 
{
  point p1, p2;

  sideline(0, shoulder,elbow, &p1,&p2);
  if( fastscan(p1,p2) ) return 0;
  sideline(1, shoulder,elbow, &p1,&p2);
  if( fastscan(p1,p2) ) return 1;
  sideline(0, elbow,wrist, &p1,&p2);
  if( fastscan(p1,p2) ) return 2;
  sideline(1, elbow,wrist, &p1,&p2);
  if( fastscan(p1,p2) ) return 3;
  sideline(0, wrist,tip, &p1,&p2);
  if( fastscan(p1,p2) ) return 4;
  sideline(1, wrist,tip, &p1,&p2);
  if( fastscan(p1,p2) ) return 5;
  return 6;
}

int evalprog(ap)  /* evaluate a prog once, affect ds,de,dw and flags */
int ap;
{
  int s1,s2;

  if( isterm(ap) ) {
    switch( ap ) {
    case SP : ds++; return 1;
    case SM : ds--; return 1;
    case EP : de++; return 1;
    case EM : de--; return 1;
    case WP : dw++; return 1;
    case WM : dw--; return 1;
    case HITQ : return hitF;
    case SEEQ : return seeF;
    case INCQ : return incF;
    case DECQ : return decF;
    case OUTQ : return outF;
    }
    return 0;
  }
  switch( cmd(ap) ) {
  case IFAND : 
    s1 = evalprog(arg1(ap));
    s2 = evalprog(arg2(ap));
    if( s1 && s2 ) return evalprog(arg3(ap)); 
    else return evalprog(arg4(ap));
  case IFOR :
    s1 = evalprog(arg1(ap));
    s2 = evalprog(arg2(ap));
    if( s1 || s2 ) return evalprog(arg3(ap));
    else return evalprog(arg4(ap));
  case IFNOT : 
    if( ! evalprog(arg1(ap)) ) return evalprog(arg2(ap));
    else return evalprog(arg3(ap));
  }
  return 0;
}

gmove(s,e,w) /* guarded move, displacement can be +/-   */
int s,e,w;
{
  int m;

  outF = 0;
  m = (s < 0) ? SM : SP;
  gmove2(m,iabs(s));
  if( finishF ) return;
  m = (e < 0) ? EM : EP;
  gmove2(m,iabs(e));
  if( finishF ) return;
  m = (w < 0) ? WM : WP;
  gmove2(m,iabs(w));
}

/* keep track of motions.  kx count no. of move, jx displacement the 
 * actual amount which is moved.  if  kx >> jx probably stuck
 * share data between gmove2() and domove()
 */
PRIVATE int ks,ke,kw, js,je,jw; 

#define emptyQ    (moveq == EMPTY)
#define isstuck   ks-iabs(js) > 5 || ke-iabs(je) > 5 || kw-iabs(jw) > 5
#define isout	  (isoutb(tip)) || (isoutb(wrist)) || (isoutb(elbow))
#define isgoal    distance(tip,target) < NearGoal
#define look()    armpos()
               
/* guarded move one joint, if hit move back and initiate
 * chain of movements to move the part that is hit away from the obstacle
 *
 * return 1 is success, 0 is failure 
 */
PRIVATE int gmove2( cmd, delta ) 
int cmd, delta;
{
  int i,m,s,moveq;
  point last;

  last = tip;
  ks = ke = kw = js = je = jw = 0;
  moveq = EMPTY;
  for( i=0; i<HUNTLIMIT; i++ ) {
    if( isgoal ) { finishF = 1; return 1;}
    if( emptyQ && delta == 0 ) return 1;
    if( isstuck ) return 0;
    if( emptyQ ) { m = cmd; delta--; } else m = moveq;
    if( jointlimit(m) ) domove(m,0); else return 0;
    if( isout ) { domove(m,1); outF = 1; return 0; }
    if( (s = ishit()) < 6 ) {
      pain += PAINSTEP;
      domove(m,1);
      switch( s ) {
      case 0 : moveq = SP; break;
      case 1 : moveq = SM; break;
      case 2 : moveq = EP; break;
      case 3 : moveq = EM; break;
      case 4 : moveq = WP; break;
      case 5 : moveq = WM; 
      }
    }else {
      moveq = EMPTY;
      if (pain > 0 ) pain--;
    }
    sumdist += distance(last,tip);
    last = tip;
  }
  return 1;
}

/* is joint within limit when execute cmd ? yes/no */
PRIVATE int jointlimit(cmd)  
int cmd;
{
  switch( cmd ) {
  case SP : return ( zshoulder < TS-1 );
  case SM : return ( zshoulder > 0 );
  case EP : return ( zelbow < TE-1 );
  case EM : return ( zelbow > 0 );
  case WP : return ( zwrist < TW-1 );
  case WM : return ( zwrist > 0 );
  }
  return 0;  /* catch all errors */
}

/* move a joint keep track of motions, dir 0 fwd, 1 back 
 * interface to a real robot by movearm()
 * and to a vision system by look(), called from gmove2()
 */
PRIVATE domove(cmd,dir) 
int cmd,dir;
{
  if( dir ) {
    switch( cmd ) {
    case SP : cmd = SM; break;
    case SM : cmd = SP; break;
    case EP : cmd = EM; break;
    case EM : cmd = EP; break;
    case WP : cmd = WM; break;
    case WM : cmd = WP;
    }
  }
  switch( cmd ) {
  case SP : ks++; js++; zshoulder++; break; 
  case SM : ks++; js--; zshoulder--; break;
  case EP : ke++; je++; zelbow++; break;
  case EM : ke++; je--; zelbow--; break;
  case WP : kw++; jw++; zwrist++; break;
  case WM : kw++; jw--; zwrist--; 
  }
  /* movearm(); */
  look();
}

/* ---- interface to the real arm and vision system ----- 

movearm(zshoulder,zelbow,zwrist)
int zshoulder,zelbow,zwrist;
{
}

look(elbow,wrist,tip)
point *elbow, *wrist, *tip;
{
  *elbow =
  *wrist =
  *tip = 
}

 -------------------------------- */
  
/* find line p3,p4 which is parallel to p1,p2 and is ARMWIDTH apart
 * dir 0 is to the left side , 1 is to the right side
 */
PRIVATE sideline(dir, p1,p2, p3,p4)
point p1, p2, *p3, *p4;
int dir;
{
  int dx, dy, nx, ny, len;

  dx = p1.x - p2.x;
  dy = p1.y - p2.y;
  len = (int) (sqrt( (double)(dx*dx + dy*dy) ) + 0.5);
  nx = ( dy == 0 ) ? 0 : -ARMWIDTH*dy/len;
  ny = ( dx == 0 ) ? 0 : ARMWIDTH*dx/len;
  if( dir ) {
    nx = -nx;
    ny = -ny;
  }
  p3->x = p1.x + nx;
  p3->y = p1.y + ny;
  p4->x = p2.x + nx;
  p4->y = p2.y + ny;
}

int distance(p1,p2)  /* dx^2 + dy^2 avoid using the expensive sqrt */
point p1,p2;
{
  int dx,dy;

  dx = p2.x - p1.x;
  dy = p2.y - p1.y;
  return ( dx*dx + dy*dy );
}


/* line plotting algorithm from
 * J.E.Bresenham, "Algorithm for computer control of a digital plotter",
 * in H. Freeman, "Tutorial and selected reading in interactive computer
 * graphics", IEEE computer society, 1980, pp.106-111, also in
 * IBM system journal, 1(4):1965.
 * 
 * the move is either straight or diagonal
 * and prepare the increment (+DL,-DL) in mx,my
 * for straight move, it is either x or y
 *
 * it steps by DL. Caution, the final point may not be exactly the end point
 * when DL != 1 usually it goes a fraction beyond the end point
 */
int fastscan(p1,p2)
point p1,p2;
{
  int x,y,dx,dy,i,k,div,mx,my,hor;

  if( isoutb(p1) || isoutb(p2) ) return 1;
  dx = p2.x - p1.x;
  dy = p2.y - p1.y;
  x = p1.x;
  y = p1.y;
  mx = (dx < 0) ? -DL : +DL;
  my = (dy < 0) ? -DL : +DL;
  dx = iabs(dx);
  dy = iabs(dy);
  if ( dx < dy ) {
    hor = 0; k = dy;
    div = dx + dx - dy;
  }else {
    hor = 1; k = dx;
    div = dy + dy - dx;
  }
  for( i=0; i<k; i+=DL ) {
    if( div < 0 ) {      /* move straight, x or y */
      if( hor ) {    
	x += mx; dx -= DL;
      }else {
	y += my; dy -= DL;
      }
      if( dx < dy ) div += dx + dx;
      else div += dy + dy;
    }else {               /* move diagonal */
      x += mx; y += my;
      dx -= DL ; dy -= DL;
      if( dx < dy ) div += dx + dx - dy - dy;
      else div += dy + dy - dx - dx;
    }
    /* do work here */
    if( im[x][y] == OBSTAC ) return 1; 
  }
  return 0;
}
